from __future__ import unicode_literals
from builtins import range
__docformat__ = 'restructuredtext'

__all__ = []

from pysparse import spmatrix
from fipy.tools import numerix

from fipy.matrices.sparseMatrix import _SparseMatrix

class _PysparseMatrix(_SparseMatrix):

    def __init__(self, matrix):
        """Creates a wrapper for a pysparse matrix

        Allows basic python operations __add__, __sub__ etc.
        Facilitate matrix populating in an easy way.

        Parameters
        ----------
        matrix : ~pysparse.spmatrix.ll_mat
            The internal Pysparse matrix
        """
        self.matrix = matrix

    def getCoupledClass(self):
        return _CoupledPysparseMeshMatrix

    def copy(self):
        return _PysparseMatrix(matrix=self.matrix.copy())

    def __getitem__(self, index):
        m = self.matrix[index]
        if isinstance(m, type(0)) or isinstance(m, type(0.)):
            return m
        else:
            return _PysparseMatrix(matrix=m)

    def __iadd__(self, other):
        self._iadd(self.matrix, other)
        return self

    def _iadd(self, L, other, sign = 1):
        if other != 0:
            L.shift(sign, other.matrix)

    def __add__(self, other):
        """
        Add two sparse matrices

            >>> L = _PysparseMatrixFromShape(rows=3, cols=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> print(L + _PysparseIdentityMatrix(size=3))
             1.000000  10.000000   3.000000  
                ---     4.141593      ---    
             2.500000      ---     1.000000  

            >>> print(L + 0)
                ---    10.000000   3.000000  
                ---     3.141593      ---    
             2.500000      ---        ---    

            >>> print(L + 3)
            Traceback (most recent call last):
            ...
            AttributeError: 'int' object has no attribute 'matrix'
        """

        if other == 0:
            return self
        else:
            L = self.matrix.copy()
            L.shift(1, other.matrix)
            return _PysparseMatrix(matrix=L)

    __radd__ = __add__

    def __sub__(self, other):

        if other == 0:
            return self
        else:
            L = self.matrix.copy()
            L.shift(-1, other.matrix)
            return _PysparseMatrix(matrix=L)

    def __rsub__(self, other):
        return -self + other

    def __isub__(self, other):
            return self._iadd(self.matrix, other, -1)

    def __mul__(self, other):
        """
        Multiply a sparse matrix by another sparse matrix

            >>> L1 = _PysparseMatrixFromShape(rows=3, cols=3)
            >>> L1.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> L2 = _PysparseIdentityMatrix(size=3)
            >>> L2.put([4.38, 12357.2, 1.1], [2, 1, 0], [1, 0, 2])

            >>> tmp = numerix.array(((1.23572000e+05, 2.31400000e+01, 3.00000000e+00),
            ...                      (3.88212887e+04, 3.14159265e+00, 0.00000000e+00),
            ...                      (2.50000000e+00, 0.00000000e+00, 2.75000000e+00)))

            >>> numerix.allclose((L1 * L2).numpyArray, tmp)
            1

        or a sparse matrix by a vector

            >>> tmp = numerix.array((29., 6.28318531, 2.5))
            >>> numerix.allclose(L1 * numerix.array((1, 2, 3), 'd'), tmp)
            1

        or a vector by a sparse matrix

            >>> tmp = numerix.array((7.5, 16.28318531,  3.))
            >>> numerix.allclose(numerix.array((1, 2, 3), 'd') * L1, tmp) ## The multiplication is broken. Numpy is calling __rmul__ for every element instead of with  the whole array.
            1


        """
        N = self.matrix.shape[1]

        if isinstance(other, _PysparseMatrix):
            return _PysparseMatrix(matrix=spmatrix.matrixmultiply(self.matrix, other.matrix))
        else:
            shape = numerix.shape(other)
            if shape == ():
                L = spmatrix.ll_mat(N, N, N)
                L.put(other * numerix.ones(N, 'l'))
                return _PysparseMatrix(matrix=spmatrix.matrixmultiply(self.matrix, L))
            elif shape == (N,):
                y = numerix.empty((self.matrix.shape[0],))
                self.matrix.matvec(other, y)
                return y
            else:
                raise TypeError

    def __rmul__(self, other):
        if isinstance(numerix.ones(1, 'l'), type(other)):
            y = other.copy()
            self.matrix.matvec_transp(other, y)
            return y
        else:
            return self * other

    @property
    def _shape(self):
        return self.matrix.shape

    @property
    def _range(self):
        return list(range(self._shape[1])), list(range(self._shape[0]))

    def put(self, vector, id1, id2):
        """
        Put elements of `vector` at positions of the matrix corresponding to (`id1`, `id2`)

            >>> L = _PysparseMatrixFromShape(rows=3, cols=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> print(L)
                ---    10.000000   3.000000  
                ---     3.141593      ---    
             2.500000      ---        ---    
        """
        self.matrix.put(vector, id1, id2)

    def putDiagonal(self, vector):
        """
        Put elements of `vector` along diagonal of matrix

            >>> L = _PysparseMatrixFromShape(rows=3, cols=3)
            >>> L.putDiagonal([3., 10., numerix.pi])
            >>> print(L)
             3.000000      ---        ---    
                ---    10.000000      ---    
                ---        ---     3.141593  
            >>> L.putDiagonal([10., 3.])
            >>> print(L)
            10.000000      ---        ---    
                ---     3.000000      ---    
                ---        ---     3.141593  
        """
        if type(vector) in [type(1), type(1.)]:
            ids = numerix.arange(self._shape[0])
            tmp = numerix.zeros((self._shape[0],), 'd')
            tmp[:] = vector
            self.put(tmp, ids, ids)
        else:
            ids = numerix.arange(len(vector))
            self.put(vector, ids, ids)

    def take(self, id1, id2):
        vector = numerix.zeros(len(id1), 'd')
        self.matrix.take(vector, id1, id2)
        return vector

    def takeDiagonal(self):
        ids = numerix.arange(self._shape[0])
        return self.take(ids, ids)

    def addAt(self, vector, id1, id2):
        """
        Add elements of `vector` to the positions in the matrix corresponding to (`id1`,`id2`)

            >>> L = _PysparseMatrixFromShape(rows=3, cols=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> L.addAt([1.73, 2.2, 8.4, 3.9, 1.23], [1, 2, 0, 0, 1], [2, 2, 0, 0, 2])
            >>> print(L)
            12.300000  10.000000   3.000000  
                ---     3.141593   2.960000  
             2.500000      ---     2.200000  
        """
        self.matrix.update_add_at(vector,
                                  numerix.asarray(id1, dtype='int32'),
                                  numerix.asarray(id2, dtype='int32'))

    def addAtDiagonal(self, vector):
        if type(vector) in [type(1), type(1.)]:
            ids = numerix.arange(self._shape[0])
            tmp = numerix.zeros((self._shape[0],), 'd')
            tmp[:] = vector
            self.addAt(tmp, ids, ids)
        else:
            ids = numerix.arange(len(vector))
            self.addAt(vector, ids, ids)

    @property
    def numpyArray(self):
        shape = self._shape
        indices = numerix.indices(shape)
        numMatrix = self.take(indices[0].ravel(), indices[1].ravel())
        return numerix.reshape(numMatrix, shape)

    def matvec(self, x):
        """
        This method is required for scipy solvers.
        """
        return self * x

    def exportMmf(self, filename):
        """
        Exports the matrix to a Matrix Market file of the given `filename`.
        """
        self.matrix.export_mtx(filename)

class _PysparseMatrixFromShape(_PysparseMatrix):

    def __init__(self, rows, cols, bandwidth=0, sizeHint=None, matrix=None, storeZeros=True):
        """Instantiates and wraps a Pysparse `ll_mat` matrix

        Parameters
        ----------
        rows : int
            The number of matrix rows
        cols : int
            The number of matrix columns
        bandwidth : int
            The proposed band width of the matrix.
        sizeHint : int
            Estimate of the number of non-zeros
        matrix : ~pysparse.spmatrix.ll_mat
            Pre-assembled Pysparse matrix to use for storage
        storeZeros : bool
            Instructs pysparse to store zero values if possible.
        """
        sizeHint = sizeHint or max(rows, cols) * bandwidth
        if matrix is None:
            tmpMatrix = spmatrix.ll_mat(1, 1, 1)
            if hasattr(tmpMatrix, 'storeZeros'):
                matrix = spmatrix.ll_mat(rows, cols, sizeHint, storeZeros)
            else:
                matrix = spmatrix.ll_mat(rows, cols, sizeHint)

        _PysparseMatrix.__init__(self, matrix=matrix)

class _PysparseMeshMatrix(_PysparseMatrixFromShape):
    def __init__(self, mesh, bandwidth=0, sizeHint=None, matrix=None, numberOfVariables=1, numberOfEquations=1, storeZeros=True):

        """Creates a `_PysparseMatrixFromShape` associated with a `Mesh`. Allows for different number of equations and/or variables

        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The `Mesh` to assemble the matrix for.
        bandwidth : int
            The proposed band width of the matrix.
        sizeHint : int
            Estimate of the number of non-zeros
        matrix : ~pysparse.spmatrix.ll_mat
            Pre-assembled Pysparse matrix to use for storage
        numberOfVariables : int
            The columns of the matrix is determined by `numberOfVariables * self.mesh.numberOfCells`.
        numberOfEquations : int
            The rows of the matrix is determined by `numberOfEquations * self.mesh.numberOfCells`.
        storeZeros : bool
            Instructs Pysparse to store zero values if possible.
        """
        self.mesh = mesh
        self.numberOfVariables = numberOfVariables
        self.numberOfEquations = numberOfEquations
        rows = numberOfEquations * self.mesh.numberOfCells
        cols = numberOfVariables * self.mesh.numberOfCells
        _PysparseMatrixFromShape.__init__(self, rows=rows, cols=cols, bandwidth=bandwidth, sizeHint=sizeHint, matrix=matrix, storeZeros=storeZeros)

    def __mul__(self, other):
        if isinstance(other, _PysparseMeshMatrix):
            return _PysparseMeshMatrix(mesh=self.mesh,
                                       matrix=spmatrix.matrixmultiply(self.matrix, other.matrix))
        else:
            return _PysparseMatrixFromShape.__mul__(self, other)

    def asTrilinosMeshMatrix(self):
        """Transforms a Pysparse matrix into a Trilinos matrix

        Maintains the Trilinos matrix as an attribute.

        Returns
        -------
        ~fipy.matrices.trilinosMatrix._TrilinosMatrix
        """
        A = self.matrix.copy()
        values, irow, jcol = A.find()

        if not hasattr(self, 'trilinosMatrix'):
            if A.shape[0] == 0:
                bandwidth = 0
            else:
                bandwidth = int(numerix.ceil(float(len(values)) / float(A.shape[0])))
            bandwidth = 1
            from fipy.matrices.trilinosMatrix import _TrilinosMeshMatrixKeepStencil
            self.trilinosMatrix = _TrilinosMeshMatrixKeepStencil(mesh=self.mesh, bandwidth=bandwidth,
                                                                 numberOfVariables=self.numberOfVariables,
                                                                 numberOfEquations=self.numberOfEquations)

        self.trilinosMatrix.addAt(values, irow, jcol)
        self.trilinosMatrix.finalize()

        return self.trilinosMatrix

    @property
    def numpyArray(self):
        from fipy.tools import parallelComm
        if parallelComm.Nproc == 1:
            return super(_PysparseMeshMatrix, self).numpyArray
        else:
            return self.asTrilinosMeshMatrix().numpyArray


    def flush(self):
        """
        Deletes the copy of the pysparse matrix held and calls `self.trilinosMatrix.flush()` if necessary.
        """

        if hasattr(self, 'trilinosMatrix'):
            if hasattr(self.matrix, 'storeZeros'):
                self.trilinosMatrix.flush(cacheStencil=self.matrix.storeZeros)
            else:
                self.trilinosMatrix.flush(cacheStencil=False)

        if (not hasattr(self, 'cache')) or (self.cache is False):
            del self.matrix

    def _test(self):
        """
        Tests

        >>> m = _PysparseMatrixFromShape(rows=3, cols=3, storeZeros=True)
        >>> m.addAt((1., 0., 2.), (0, 2, 1), (1, 2, 0))
        >>> print(not hasattr(m.matrix, 'storeZeros') or numerix.allequal(list(m.matrix.keys()), [(0, 1), (1, 0), (2, 2)]))
        True
        >>> print(not hasattr(m.matrix, 'storeZeros') or numerix.allequal(list(m.matrix.values()), [1., 2., 0.]))
        True
        >>> m = _PysparseMatrixFromShape(rows=3, cols=3, storeZeros=False)
        >>> m.addAt((1., 0., 2.), (0, 2, 1), (1, 2, 0))
        >>> print(numerix.allequal(list(m.matrix.keys()), [(0, 1), (1, 0)]))
        True
        >>> print(numerix.allequal(list(m.matrix.values()), numerix.array([1.0, 2.0])))
        True

        """
        pass

class _PysparseIdentityMatrix(_PysparseMatrixFromShape):
    """
    Represents a sparse identity matrix for pysparse.
    """
    def __init__(self, size):
        """Create a sparse matrix with `1` in the diagonal

            >>> print(_PysparseIdentityMatrix(size=3))
             1.000000      ---        ---    
                ---     1.000000      ---    
                ---        ---     1.000000  
        """
        _PysparseMatrixFromShape.__init__(self, rows=size, cols=size, bandwidth = 1)
        ids = numerix.arange(size)
        self.put(numerix.ones(size, 'd'), ids, ids)

class _PysparseIdentityMeshMatrix(_PysparseIdentityMatrix):
    def __init__(self, mesh):
        """Create a sparse matrix associated with a `Mesh` with `1` in the diagonal

            >>> from fipy import Grid1D
            >>> from fipy.tools import serialComm
            >>> mesh = Grid1D(nx=3, communicator=serialComm)
            >>> print(_PysparseIdentityMeshMatrix(mesh=mesh))
             1.000000      ---        ---    
                ---     1.000000      ---    
                ---        ---     1.000000  
        """
        _PysparseIdentityMatrix.__init__(self, size=mesh.numberOfCells)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()






from __future__ import unicode_literals
from builtins import object
from builtins import range
from builtins import zip
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import numerix

class _UniformOrigin(object):
    """
    Used to calculate the origin for uniform grids in a
    dimensionally-independent way.
    """

    @staticmethod
    def calcOrigin(origin, offset, ds, scale):
        newOrigin  = PhysicalField(value=origin)
        newOrigin /= scale

        if type(offset) in [int, float]:
            newOrigin += offset * ds[0]
        else:
            newOrigin += [[o*float(d)] for o, d in zip(offset, ds)]

        return newOrigin

class _DOffsets(object):
    """
    For use by non-uniform grid builders.
    """

    @staticmethod
    def calcDOffsets(ds, ns, offset):
        """
        Parameters
        ----------
        ds : list
            Spacing in each grid direction, e.g. `[dx, dy]`
        ns : list
            Number of grid spacings in each direction, e.g. `[nx, ny]`
        offset : list
            Displacement of grid spacings, e.g., `[Ox, Oy]`

        Returns
        -------
        offsetList : list
            Contains the analogous scalars to `XOffset`, `YOffset`, and
            `ZOffset`, whichever are applicable for the dimensionality.
        newDs : list
            proper `[dx, [dy, ...]]` values
        """
        offsetList = []
        newDs = []

        if type(offset) in [int, float]:
            offset = [offset]

        for d, n, i in zip(ds, ns, list(range(len(ds)))):
            if numerix.getShape(d) is not ():
                offsetList.append(numerix.sum(d[0:offset[i]]))
                newDs.append(d[offset[i]:offset[i] + n])
            else:
                if len(offset) == 1:
                    offsetList.append(d * offset[0])
                else:
                    offsetList.append(d * offset[i])

                newDs.append(d)

        return offsetList, newDs

class _AbstractNumPts(object):
    """
    Interface definition for `NumPts` calculators.
    """

    @staticmethod
    def calcNs(ns, ds):
        raise NotImplementedError

class _NonuniformNumPts(_AbstractNumPts):
    """
    For use by non-uniform grid builders.
    """

    @staticmethod
    def calcNs(ns, ds):
        axis = ["x", "y", "z"][:len(ns)]
        newNs = []

        for a, d, n in zip(axis, ds, ns):
            newNs.append(_NonuniformNumPts._calcNumPts(d=d, n=n, axis=a))

        return newNs

    @staticmethod
    def _calcNumPts(d, n = None, axis = "x"):
        """
        Calculate the number of cells along the specified axis, based
        on either the specified number or on the number elements in the
        cell  `d` spacings.

        Used by the `Grid` meshes.

        This tests a bug that was occurring with `PeriodicGrid1D` when
        using a numpy float as the argument for the grid spacing.

           >>> from fipy.meshes.periodicGrid1D import PeriodicGrid1D
           >>> PeriodicGrid1D(nx=2, dx=numerix.float32(1.))
           PeriodicGrid1D(dx=1.0, nx=2)
        """

        if type(d) in [int, float] or numerix.shape(d) == ():
            n = int(n or 1)
        else:
            n = int(n or len(d))
            if n != len(d) and len(d) != 1:
                raise IndexError("n%s != len(d%s)" % (axis, axis))

        return n

class _UniformNumPts(_AbstractNumPts):
    """
    For use by uniform grid builders.
    """

    @staticmethod
    def calcNs(ns, ds):
        return [int(x) for x in ns]

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

from fipy.boundaryConditions.boundaryCondition import BoundaryCondition
from fipy.boundaryConditions.fixedValue import FixedValue
from fipy.tools import vector

__all__ = ["FixedFlux"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class FixedFlux(BoundaryCondition):
    r"""

    The `FixedFlux` boundary condition adds a contribution, equivalent to a
    fixed flux (Neumann condition), to the equation's RHS vector.  The
    contribution, given by `value`, is only added to entries corresponding to
    the specified `faces`, and is weighted by the face areas.

    """

    def __init__(self, faces, value):
        """
        Parameters
        ----------
        faces : :obj:`~fipy.variables.faceVariable.FaceVariable` of :obj:`bool`
            Mask of faces where this condition applies.
        value : float
            Value to impose.
        """
        BoundaryCondition.__init__(self, faces, value)
        ## The extra index [self.faces.value] makes self.contribution the same length as self.adjacentCellIDs
        self.contribution = (self.value * self.faces.mesh._faceAreas)[self.faces.value]

    def _buildMatrix(self, SparseMatrix, Ncells, MaxFaces, coeff):
        """Leave **L** unchanged and add gradient to **b**

        Parameters
        ----------
        SparseMatrix : ~fipy.matrices.sparseMatrix._SparseMatrix
            *unused* (empty matrix)
        Ncells : int
            Size of **b** vector
        MaxFaces
            *unused*
        coeff : list
            *unused*
        """

        bb = numerix.zeros((Ncells,), 'd')

        if not self.boundaryConditionApplied:
            vector.putAdd(bb, self.adjacentCellIDs, -self.contribution)
            self.boundaryConditionApplied = True

        return (0, bb)

    def _getDerivative(self, order):
        if order == 1:
            return FixedValue(self.faces, self.value)
        else:
            return BoundaryCondition._getDerivative(self, order)

    def _test(self):
        """
        The following tests check that `self.contributions` is the same length as
        `self.adjacentCellIDs`.

           >>> from fipy import *
           >>> m = Grid1D(nx = 10)
           >>> v = FaceVariable(mesh=m)
           >>> bc = FixedFlux(value=v.globalValue[-1], faces=m.facesRight)
           >>> len(bc.contribution) == len(bc.adjacentCellIDs)
           True
           >>> bc = FixedFlux(value=v, faces=m.facesRight)
           >>> len(bc.contribution) == len(bc.adjacentCellIDs)
           True
        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
from pysparse.precon import precon

from fipy.solvers.pysparse.preconditioners.preconditioner import Preconditioner

__all__ = ["JacobiPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class JacobiPreconditioner(Preconditioner):
    """
    Jacobi preconditioner for Pysparse.
    Really just a wrapper class for `pysparse.precon.jacobi`.
    """
    def _applyToMatrix(self, A):
        """
        Returns (preconditioning matrix, resulting matrix)
        """
        return precon.jacobi(A), A.to_csr()



from __future__ import print_function
from __future__ import unicode_literals
from builtins import input
import os
from distutils.core import Command
from future.utils import text_to_native_str

from ._nativize import nativize_all

__all__ = [text_to_native_str("copy_script")]

class copy_script(Command):
    description = "copy an example script into a new editable file"

    # List of option tuples: long name, short name (None if no short
    # name), and help string.
    user_options = [
        # Select installation scheme and set base director(y|ies)
        ('From=', None,
         "path and file name containing script to copy"),
        ('To=', None,
         "path and file name to save script to")
     ]
    user_options = [nativize_all(u) for u in user_options]

    def initialize_options(self):
        self.From = None
        self.To = None

    def finalize_options(self):
        if self.From == None:
            raise SyntaxError("Please specify a '--From' input script file")

        if self.To == None:
            raise SyntaxError("Please specify a '--To' output script file")

        if os.path.exists(os.path.expanduser(self.To)):
            ans = "junk"

            while (len(ans) > 0) and ("yes".find(ans.lower()) is not 0) and ("no".find(ans.lower()) is not 0):
                ans = input("The file '%s' already exists. Overwrite? [n] "%self.To)

            if ans is '':
                ans = 'no'

            if ("no".find(ans.lower()) is 0):
                self.To = input("Please give a name for the ouput file: ")
                self.finalize_options()

    def run(self):
        import imp
        import fipy.tests.doctestPlus

        mod = imp.load_source("copy_script_module", self.From)
        script = fipy.tests.doctestPlus._getScript(name = "copy_script_module")
        script = "\n\n## This script was derived from\n## '%s'\n\n%s"%(self.From, script)
        with open(self.To, "w") as f:
            f.write(script)

        print("Script code exported from '%s' to '%s'"%(self.From, self.To))



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import ML

from fipy.solvers.trilinos.preconditioners.preconditioner import Preconditioner

__all__ = ["MultilevelSolverSmootherPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MultilevelSolverSmootherPreconditioner(Preconditioner):
    """
    Multilevel preconditioner for Trilinos solvers using Aztec solvers
    as smoothers.

    """
    def __init__(self, levels=10):
        """
        Initialize the multilevel preconditioner

        - `levels`: Maximum number of levels
        """
        self.levels = levels

    def _applyToSolver(self, solver, matrix):
        if matrix.NumGlobalNonzeros() <= matrix.NumGlobalRows():
            return

        self.Prec = ML.MultiLevelPreconditioner(matrix, False)
        self.Prec.SetParameterList({text_to_native_str("output"): 0, text_to_native_str("smoother: type") : text_to_native_str("Aztec"), text_to_native_str("smoother: Aztec as solver") : True})
        self.Prec.ComputePreconditioner()
        solver.SetPrecOperator(self.Prec)



from __future__ import unicode_literals
from builtins import zip
from builtins import str
from builtins import range
__docformat__ = 'restructuredtext'

__all__ = []

import dis
import sys

from fipy.variables.variable import Variable
from fipy.tools import numerix

def _OperatorVariableClass(baseClass=object):
    class _OperatorVariable(baseClass):
        def __init__(self, op, var, opShape=(), canInline=True, unit=None, inlineComment=None, valueMattersForUnit=None, *args, **kwargs):
            self.op = op
            self.var = var
            self.opShape = opShape
            self._unit = unit
            if valueMattersForUnit is None:
                self.valueMattersForUnit = [False for v in var]
            else:
                self.valueMattersForUnit = valueMattersForUnit
            self.canInline = canInline  #allows for certain functions to opt out of --inline
            baseClass.__init__(self, value=None, *args, **kwargs)
            self.name = ''
            for var in self.var:    #C does not accept units
                if not var.unit.isDimensionless():
                    self.canInline = False
                    break

            for aVar in self.var:
                self._requires(aVar)

            self.dontCacheMe()

            self.comment = inlineComment

        def __setitem__(self, index, value):
            raise TypeError("The value of an `_OperatorVariable` cannot be assigned")

        def setValue(self, value, unit=None, where=None):
            raise TypeError("The value of an `_OperatorVariable` cannot be assigned")

        def _calcValue(self):
            if not self.canInline:
                return self._calcValue_()
            else:
                from fipy.tools import inline
                if inline.doInline:
                    return self._execInline(comment=self.comment)
                else:
                    return self._calcValue_()

        def _calcValue_(self):
            pass

        def _isCached(self):
            return (Variable._isCached(self)
                    or (len(self.subscribedVariables) > 1 and not self._cacheNever))

        def _getCstring(self, argDict={}, id="", freshen=False):
            if self.canInline: # and not self._isCached():
                s = self._getRepresentation(style="C", argDict=argDict, id=id, freshen=freshen)
            else:
                s = baseClass._getCstring(self, argDict=argDict, id=id)
            if freshen:
                self._markFresh()

            return s

        def _getRepresentation(self, style="__repr__", argDict={}, id=id, freshen=False):
            """

            Parameters
            ----------
            style : {'__repr__', 'name', 'TeX', 'C'}
               desired formatting for representation
            """
            if isinstance(self.op, numerix.ufunc):
                return "%s(%s)" % (self.op.__name__, ", ".join([self.__var(i, style, argDict, id, freshen)
                                                               for i in range(len(self.var))]))

            try:
                instructions = dis.get_instructions(self.op.__code__)
                parseInstructions = self._py3kInstructions
            except AttributeError:
                instructions = [ord(byte) for byte in self.op.__code__.co_code]
                parseInstructions = self._py2kInstructions

            return parseInstructions(instructions, style=style, argDict=argDict, id=id, freshen=freshen)

        def __var(self, i, style, argDict, id, freshen):
            v = self.var[i]
            if style == "__repr__":
                result = repr(v)
            elif style == "name":
                if isinstance(v, Variable):
                    result = v.name
                    if len(result) == 0:
                        # The string form of a variable
                        # would probably be too long and messy.
                        # Just give shorthand.
                        result = "%s(...)" % v.__class__.__name__
                elif type(v) in (type(1), type(1.)):
                    result = repr(v)
                else:
                    # The string form of anything but a
                    # number would be too long and messy.
                    # Just give shorthand.
                    result = "<...>"

            elif style == "TeX":
                raise Exception("TeX style not yet implemented")
            elif style == "C":
                if not v._isCached():
                    result = v._getCstring(argDict, id=id + str(i), freshen=freshen)
                    if isinstance(v, Variable):
                        v._value = None
                    else:
                        v.value = None
                else:
                    result = v._variableClass._getCstring(v, argDict,
                                                               id=id + str(i),
                                                               freshen=False)
            else:
                raise SyntaxError("Unknown style: %s" % style)

            return result

        _unop = {
            10: "+", 11: "-", 12: "not ", 15: "~"
        }

        _binop = {
            19: "**", 20: "*", 21: "/", 22: "%", 23: "+", 24: "-", 26: "//", 27: "/",
                    62: "<<", 63: ">>", 64: "&", 65: "^", 66: "|", 106: "=="
        }

        def _py2kInstructions(self, bytecodes, style, argDict, id, freshen):
            def _popIndex():
                return bytecodes.pop(0) + bytecodes.pop(0) * 256

            allbytecodes = bytecodes[:]

            stack = []

            while len(bytecodes) > 0:
                bytecode = bytecodes.pop(0)
                if dis.opname[bytecode] == 'UNARY_CONVERT':
                    stack.append("`" + stack.pop() + "`")
                elif dis.opname[bytecode] == 'BINARY_SUBSCR':
                    stack.append(stack.pop(-2) + "[" + stack.pop() + "]")
                elif dis.opname[bytecode] == 'RETURN_VALUE':
                    s = stack.pop()
                    if style == 'C':
                        return s.replace('numerix.', '').replace('arc', 'a')
                    else:
                        return s
                elif dis.opname[bytecode] == 'LOAD_CONST':
                    stack.append(self.op.__code__.co_consts[_popIndex()])
                elif dis.opname[bytecode] == 'LOAD_ATTR':
                    stack.append(stack.pop() + "." + self.op.__code__.co_names[_popIndex()])
                elif dis.opname[bytecode] == 'COMPARE_OP':
                    stack.append(stack.pop(-2) + " " + dis.cmp_op[_popIndex()] + " " + stack.pop())
                elif dis.opname[bytecode] == 'LOAD_GLOBAL':
                    counter = _popIndex()
                    stack.append(self.op.__code__.co_names[counter])
                elif dis.opname[bytecode] == 'LOAD_FAST':
                    stack.append(self.__var(_popIndex(), style=style, argDict=argDict, id=id, freshen=freshen))
                elif dis.opname[bytecode] == 'CALL_FUNCTION':
                    args = []
                    for j in range(bytecodes.pop(1)):
                        # keyword parameters
                        args.insert(0, stack.pop(-2) + " = " + stack.pop())
                    for j in range(bytecodes.pop(0)):
                        # positional parameters
                        args.insert(0, stack.pop())
                    stack.append(stack.pop() + "(" + ", ".join(args) + ")")
                elif dis.opname[bytecode] == 'LOAD_DEREF':
                    free = self.op.__code__.co_cellvars + self.op.__code__.co_freevars
                    stack.append(free[_popIndex()])
                elif bytecode in self._unop:
                    stack.append(self._unop[bytecode] + '(' + stack.pop() + ')')
                elif bytecode in self._binop:
                    stack.append(stack.pop(-2) + " " + self._binop[bytecode] + " " + stack.pop())
                else:
                    raise SyntaxError("Unknown bytecode: %s in %s of %s, got %s" % (
                       repr(bytecode),
                       repr([bytecode] + bytecodes),
                       repr(allbytecodes),
                       repr(stack)))

        def _py3kInstructions(self, instructions, style, argDict, id, freshen):
            stack = []
            
            for ins in instructions:
                if ins.opname == 'UNARY_CONVERT':
                    stack.append("`" + stack.pop() + "`")
                elif ins.opname == 'BINARY_SUBSCR':
                    stack.append(stack.pop(-2) + "[" + stack.pop() + "]")
                elif ins.opname == 'RETURN_VALUE':
                    s = stack.pop()
                    if style == 'C':
                        return s.replace('numerix.', '').replace('arc', 'a')
                    else:
                        return s
                elif ins.opname == 'LOAD_CONST':
                    stack.append(ins.argval)
                elif ins.opname == 'LOAD_ATTR':
                    stack.append(stack.pop() + "." + ins.argval)
                elif ins.opname == 'COMPARE_OP':
                    stack.append(stack.pop(-2) + " " + dis.cmp_op[ins.arg] + " " + stack.pop())
                elif ins.opname == 'LOAD_GLOBAL':
                    stack.append(ins.argval)
                elif ins.opname == 'LOAD_FAST':
                    stack.append(self.__var(ins.arg, style=style, argDict=argDict, id=id, freshen=freshen))
                elif ins.opname == 'CALL_FUNCTION':
                    # args are last ins.arg items on stack
                    args, stack = stack[-ins.arg:], stack[:-ins.arg]
                    stack.append(stack.pop() + "(" + ", ".join(args) + ")")
                elif ins.opname == 'CALL_FUNCTION_KW':
                    kws = list(stack.pop())
                    # args are last ins.arg items on stack
                    args, stack = stack[-ins.arg:], stack[:-ins.arg]
                    kwargs = []
                    while kws:
                        kwargs.append(kws.pop() + "=" + args.pop())
                    stack.append(stack.pop() + "(" + ", ".join(args + kwargs) + ")")
                elif ins.opname == 'LOAD_DEREF':
                    stack.append(ins.argval)
                elif ins.opcode in self._unop:
                    stack.append(self._unop[ins.opcode] + '(' + stack.pop() + ')')
                elif ins.opcode in self._binop:
                    stack.append(stack.pop(-2) + " " + self._binop[ins.opcode] + " " + stack.pop())
                else:
                    raise SyntaxError("Unknown instruction: %s" % repr(ins))

        @property
        def _varProxy(self):
            """list of dimensional scalars that stand in for `self.var`

            Used for determining units of result without doing expensive computation"""

            return [v if valueMatters else v._unitAsOne for v, valueMatters in zip(self.var, self.valueMattersForUnit)]

        def __repr__(self):
            return self._getRepresentation()

        def __reduce__(self):
            """
            Allows `_OperatorVariables` to be pickled
            """
            state =  self.__getstate__()
            if 'mesh' in list(state.keys()):
                args = (state['mesh'],)
            else:
                args = ()

            return (self._variableClass, args, self.__getstate__())

        def _getName(self):
            name = baseClass._getName(self)
            if len(name) == 0:
                name = self._getRepresentation(style="name")
            return name

        name = property(_getName, baseClass._setName)

        @property
        def shape(self):
            if self.opShape is not None:
                return self.opShape
            else:
                return baseClass.getShape(self)
##             return baseClass.getShape(self) or self.opShape

    return _OperatorVariable

def _testBinOp(self):
    """
    Test of `_getRepresentation`

        >>> v1 = Variable((1, 2, 3, 4))
        >>> v2 = Variable((5, 6, 7, 8))
        >>> v3 = Variable((9, 10, 11, 12))
        >>> v4 = Variable((13, 14, 15, 16))

        >>> from future.utils import text_to_native_str as ttns

        >>> ttns((v1 * v2)._getRepresentation())
        '(Variable(value=array([1, 2, 3, 4])) * Variable(value=array([5, 6, 7, 8])))'

        >>> ttns((v1 * v2)._getRepresentation(style='C', id=""))
        '(var0[i] * var1[i])'

        >>> ttns((v1 * v2 + v3 * v4)._getRepresentation(style='C', id=""))
        '((var00[i] * var01[i]) + (var10[i] * var11[i]))'

        >>> ttns((v1 - v2)._getRepresentation(style='C', id=""))
        '(var0[i] - var1[i])'

        >>> ttns((v1 / v2)._getRepresentation(style='C', id=""))
        '(var0[i] / var1[i])'

        >>> ttns((v1 - 1)._getRepresentation(style='C', id=""))
        '(var0[i] - var1)'

        >>> ttns((5 * v2)._getRepresentation(style='C', id=""))
        '(var0[i] * var1)'

        >>> ttns((v1 / v2 - v3 * v4 + v1 * v4)._getRepresentation(style='C', id=""))
        '(((var000[i] / var001[i]) - (var010[i] * var011[i])) + (var10[i] * var11[i]))'

    Check that unit works for a `binOp`

        >>> (Variable(value="1 m") * Variable(value="1 s")).unit == Variable(value="1 s*m").unit
        True

        >>> (Variable(value="1 m") / Variable(value="0 s")).unit
        <PhysicalUnit m/s>

        >>> a = -(numerix.sin(Variable() * Variable()))

    Check that `getTypeCode()` works as expected.

        >>> a = Variable(1.) * Variable(1)
        >>> print(a.getsctype() == numerix.float64)
        1

    The following test is to correct an `--inline` bug that was
    being thrown by the Cahn-Hilliard example. The fix for this
    bug was to add the last line to the following code in
    `_getRepresentation()`.

        >>> ##elif style == "C":
        >>> ##    counter = _popIndex()
        >>> ##    if not self.var[counter]._isCached():
        >>> ##        stack.append(self.var[counter]._getCstring(argDict, id=id + str(counter), freshen=freshen))
        >>> ##        self.var[counter].value=None

    This is the test that fails if the last line above is removed
    from `_getRepresentation()`, the `binOp.value` statement
    below will return `1.0` and not `0.5`.

        >>> from fipy import numerix
        >>> def doBCs(binOp):
        ...     unOp1 = -binOp
        ...     print(binOp.value)
        >>> var = Variable(1.)
        >>> binOp = 1. * var
        >>> unOp = -binOp
        >>> print(unOp.value)
        -1.0
        >>> doBCs(binOp)
        1.0
        >>> var.value = (0.5)
        >>> print(unOp.value)
        -0.5
        >>> unOp2 = -binOp
        >>> print(binOp.value)
        0.5

        >>> from fipy.variables.cellVariable import CellVariable
        >>> from fipy.variables.faceVariable import FaceVariable

        >>> from fipy.meshes import Grid2D
        >>> mesh = Grid2D(nx=3)


    `CellVariable` * `CellVariable`

        >>> cv = CellVariable(mesh=mesh, value=(0, 1, 2))
        >>> cvXcv = cv * cv
        >>> print(cvXcv)
        [0 1 4]
        >>> print(isinstance(cvXcv, CellVariable))
        1

    `CellVariable` * `FaceVariable`

        >>> fv = FaceVariable(mesh=mesh, value=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
        >>> fvXcv = fv * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> cvXfv = cv * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    rank-0 `CellVariable` * rank-1 `CellVariable`

        >>> vcv = CellVariable(mesh=mesh, value=((0, 1, 2), (1, 2, 3)), rank=1)
        >>> vcvXcv = vcv * cv
        >>> print(vcvXcv)
        [[0 1 4]
         [0 2 6]]
        >>> print(isinstance(vcvXcv, CellVariable))
        1
        >>> print(vcvXcv.rank)
        1
        >>> cvXvcv = cv * vcv
        >>> print(cvXvcv)
        [[0 1 4]
         [0 2 6]]
        >>> print(isinstance(cvXvcv, CellVariable))
        1
        >>> print(cvXvcv.rank)
        1

    `CellVariable` * rank-1 `FaceVariable`

        >>> vfv = FaceVariable(mesh=mesh, value=((0, 1, 2, 3, 1, 2, 3, 6, 2, 1),
        ...                                      (1, 2, 3, 4, 3, 4, 5, 9, 6, 3)), rank=1)
        >>> vfvXcv = vfv * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> cvXvfv = cv * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    `CellVariable` * Scalar

        >>> cvXs = cv * 3
        >>> print(cvXs)
        [0 3 6]
        >>> print(isinstance(cvXs, CellVariable))
        1
        >>> sXcv = 3 * cv
        >>> print(sXcv)
        [0 3 6]
        >>> print(isinstance(sXcv, CellVariable))
        1

    `CellVariable` * Vector

        >>> cvXv2 = cv * (3, 2)
        >>> print(cvXv2)
        [[0 3 6]
         [0 2 4]]
        >>> cvXv2 = cv * [[3], [2]]
        >>> print(cvXv2)
        [[0 3 6]
         [0 2 4]]
        >>> print(isinstance(cvXv2, CellVariable))
        1
        >>> print(cvXv2.rank)
        1
        >>> v2Xcv = (3, 2) * cv
        >>> print(v2Xcv)
        [[0 3 6]
         [0 2 4]]
        >>> v2Xcv = [[3], [2]] * cv
        >>> print(v2Xcv)
        [[0 3 6]
         [0 2 4]]
        >>> print(isinstance(v2Xcv, CellVariable))
        1
        >>> print(v2Xcv.rank)
        1

        >>> cvXv3 = cv * (3, 2, 1)
        >>> print(cvXv3)
        [0 2 2]
        >>> print(isinstance(cvXv3, CellVariable))
        1
        >>> v3Xcv = (3, 2, 1) * cv
        >>> print(v3Xcv)
        [0 2 2]
        >>> print(isinstance(v3Xcv, CellVariable))
        1

        >>> cvXv4 = cv * (3, 2, 1, 0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        TypeError: can't multiply sequence to non-int
        >>> v4Xcv = (3, 2, 1, 0) * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        TypeError: can't multiply sequence to non-int


    `CellVariable` * `Variable` Scalar

        >>> cvXsv = cv * Variable(value=3)
        >>> print(cvXsv)
        [0 3 6]
        >>> print(isinstance(cvXsv, CellVariable))
        1
        >>> svXcv = Variable(value=3) * cv
        >>> print(svXcv)
        [0 3 6]
        >>> print(isinstance(svXcv, CellVariable))
        1

    `binOp` `CellVariable` * `binOp` `Variable` Scalar

        >>> cvcvXsvsv = (cv * cv) * (Variable(value=3) * Variable(value=3))
        >>> print(cvcvXsvsv)
        [ 0  9 36]
        >>> print(isinstance(cvcvXsvsv, CellVariable))
        1
        >>> svsvXcvcv = (Variable(value=3) * Variable(value=3)) * (cv * cv)
        >>> print(svsvXcvcv)
        [ 0  9 36]
        >>> print(isinstance(svsvXcvcv, CellVariable))
        1

    `CellVariable` * `Variable` Vector

        >>> cvXv2v = cv * Variable(value=(3, 2))
        >>> print(cvXv2v)
        [[0 3 6]
         [0 2 4]]
        >>> cvXv2v = cv * Variable(value=((3,), (2,)))
        >>> print(cvXv2v)
        [[0 3 6]
         [0 2 4]]
        >>> print(isinstance(cvXv2v, CellVariable))
        1
        >>> print(cvXv2v.rank)
        1
        >>> v2vXcv = Variable(value=(3, 2)) * cv
        >>> v2vXcv = Variable(value=((3,), (2,))) * cv
        >>> print(v2vXcv)
        [[0 3 6]
         [0 2 4]]
        >>> print(isinstance(v2vXcv, CellVariable))
        1
        >>> print(v2vXcv.rank)
        1

        >>> cvXv3v = cv * Variable(value=(3, 2, 1))
        >>> print(cvXv3v)
        [0 2 2]
        >>> print(isinstance(cvXv3v, CellVariable))
        1
        >>> v3vXcv = Variable(value=(3, 2, 1)) * cv
        >>> print(v3vXcv)
        [0 2 2]
        >>> print(isinstance(v3vXcv, CellVariable))
        1

        >>> cvXv4v = cv * Variable(value=(3, 2, 1, 0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4vXcv = Variable(value=(3, 2, 1, 0)) * cv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


    `CellVariable` * `CellGradVariable`

        >>> cvXcgv = cv * cv.grad
        >>> print(cvXcgv)
        [[ 0.  1.  1.]
         [ 0.  0.  0.]]
        >>> print(isinstance(cvXcgv, CellVariable))
        1
        >>> print(cvXcgv.rank)
        1

    `FaceVariable` * `FaceVariable`

        >>> fvXfv = fv * fv
        >>> print(fvXfv)
        [ 0  1  4  9 16 25 36 49 64 81]
        >>> print(isinstance(fvXfv, FaceVariable))
        1

    `FaceVariable` * rank-1 `CellVariable`

        >>> vcvXfv = vcv * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> fvXvcv = fv * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    `FaceVariable` * rank-1 `FaceVariable`

        >>> vfvXfv = vfv * fv
        >>> print(vfvXfv)
        [[ 0  1  4  9  4 10 18 42 16  9]
         [ 0  2  6 12 12 20 30 63 48 27]]
        >>> print(isinstance(vfvXfv, FaceVariable))
        1
        >>> print(vfvXfv.rank)
        1
        >>> fvXvfv = fv * vfv
        >>> print(fvXvfv)
        [[ 0  1  4  9  4 10 18 42 16  9]
         [ 0  2  6 12 12 20 30 63 48 27]]
        >>> print(isinstance(fvXvfv, FaceVariable))
        1
        >>> print(fvXvfv.rank)
        1

    `FaceVariable` * Scalar

        >>> fvXs = fv * 3
        >>> print(fvXs)
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print(isinstance(fvXs, FaceVariable))
        1
        >>> sXfv = 3 * fv
        >>> print(sXfv)
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print(isinstance(sXfv, FaceVariable))
        1

    `FaceVariable` * Vector

        >>> fvXv2 = fv * (3, 2)
        >>> print(fvXv2)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> fvXv2 = fv * ((3,), (2,))
        >>> print(fvXv2)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print(isinstance(fvXv2, FaceVariable))
        1
        >>> print(fvXv2.rank)
        1
        >>> v2Xfv = (3, 2) * fv
        >>> print(v2Xfv)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> v2Xfv = ((3,), (2,)) * fv
        >>> print(v2Xfv)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print(isinstance(v2Xfv, FaceVariable))
        1
        >>> print(v2Xfv.rank)
        1

        >>> fvXv3 = fv * (3, 2, 1) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3Xfv = (3, 2, 1) * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

        >>> fvXv10 = fv * (9, 8, 7, 6, 5, 4, 3, 2, 1, 0)
        >>> print(fvXv10)
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print(isinstance(fvXv10, FaceVariable))
        1
        >>> v10Xfv = (9, 8, 7, 6, 5, 4, 3, 2, 1, 0) * fv
        >>> print(v10Xfv)
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print(isinstance(v10Xfv, FaceVariable))
        1

    `FaceVariable` * `Variable` Scalar

        >>> fvXsv = fv * Variable(value=3)
        >>> print(fvXsv)
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print(isinstance(fvXsv, FaceVariable))
        1
        >>> svXfv = Variable(value=3) * fv
        >>> print(svXfv)
        [ 0  3  6  9 12 15 18 21 24 27]
        >>> print(isinstance(svXfv, FaceVariable))
        1

    `FaceVariable` * `Variable` Vector

        >>> fvXv2v = fv * Variable(value=(3, 2))
        >>> print(fvXv2v)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> fvXv2v = fv * Variable(value=((3,), (2,)))
        >>> print(fvXv2v)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print(isinstance(fvXv2v, FaceVariable))
        1
        >>> print(fvXv2v.rank)
        1
        >>> v2vXfv = Variable(value=(3, 2)) * fv
        >>> print(v2vXfv)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> v2vXfv = Variable(value=((3,), (2,))) * fv
        >>> print(v2vXfv)
        [[ 0  3  6  9 12 15 18 21 24 27]
         [ 0  2  4  6  8 10 12 14 16 18]]
        >>> print(isinstance(v2vXfv, FaceVariable))
        1
        >>> print(v2vXfv.rank)
        1

        >>> fvXv3v = fv * Variable(value=(3, 2, 1)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXfv = Variable(value=(3, 2, 1)) * fv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

        >>> fvXv10v = fv * Variable(value=(9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
        >>> print(fvXv10v)
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print(isinstance(fvXv10v, FaceVariable))
        1
        >>> v10vXfv = Variable(value=(9, 8, 7, 6, 5, 4, 3, 2, 1, 0)) * fv
        >>> print(v10vXfv)
        [ 0  8 14 18 20 20 18 14  8  0]
        >>> print(isinstance(v10vXfv, FaceVariable))
        1



    rank-1 `CellVariable` * rank-1 `CellVariable`

        >>> vcvXvcv = vcv * vcv
        >>> print(vcvXvcv)
        [[0 1 4]
         [1 4 9]]
        >>> print(isinstance(vcvXvcv, CellVariable))
        1
        >>> print(vcvXvcv.rank)
        1

    rank-1 `CellVariable` * rank-1 `FaceVariable`

        >>> vfvXvcv = vfv * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> vcvXvfv = vcv * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    rank-1 `CellVariable` * Scalar

        >>> vcvXs = vcv * 3
        >>> print(vcvXs)
        [[0 3 6]
         [3 6 9]]
        >>> print(isinstance(vcvXs, CellVariable))
        1
        >>> print(vcvXs.rank)
        1
        >>> sXvcv = 3 * vcv
        >>> print(sXvcv)
        [[0 3 6]
         [3 6 9]]
        >>> print(isinstance(vcvXs, CellVariable))
        1
        >>> print(vcvXs.rank)
        1

    rank-1 `CellVariable` * Vector

        >>> vcvXv2 = vcv * (3, 2)
        >>> print(vcvXv2)
        [[0 3 6]
         [2 4 6]]
        >>> vcvXv2 = vcv * ((3,), (2,))
        >>> print(vcvXv2)
        [[0 3 6]
         [2 4 6]]
        >>> print(isinstance(vcvXv2, CellVariable))
        1
        >>> print(vcvXv2.rank)
        1
        >>> v2Xvcv = (3, 2) * vcv
        >>> print(v2Xvcv)
        [[0 3 6]
         [2 4 6]]
        >>> v2Xvcv = ((3,), (2,)) * vcv
        >>> print(v2Xvcv)
        [[0 3 6]
         [2 4 6]]
        >>> print(isinstance(v2Xvcv, CellVariable))
        1
        >>> print(v2Xvcv.rank)
        1

        >>> vcvXv3 = vcv * (3, 2, 1)
        >>> print(vcvXv3)
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(vcvXv3, CellVariable)
        1
        >>> print(vcvXv3.rank)
        1
        >>> v3Xvcv = (3, 2, 1) * vcv
        >>> print(v3Xvcv)
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(v3Xvcv, CellVariable)
        1
        >>> print(v3Xvcv.rank)
        1

        >>> vcvXv4 = vcv * (3, 2, 1, 0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4Xvcv = (3, 2, 1, 0) * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    rank-1 `CellVariable` * `Variable` Scalar

        >>> vcvXsv = vcv * Variable(value=3)
        >>> print(vcvXsv)
        [[0 3 6]
         [3 6 9]]
        >>> print(isinstance(vcvXsv, CellVariable))
        1
        >>> print(vcvXsv.rank)
        1
        >>> svXvcv = Variable(value=3) * vcv
        >>> print(svXvcv)
        [[0 3 6]
         [3 6 9]]
        >>> print(isinstance(svXvcv, CellVariable))
        1
        >>> print(svXvcv.rank)
        1

    rank-1 `CellVariable` * `Variable` Vector

        >>> vcvXv2v = vcv * Variable(value=(3, 2))
        >>> print(vcvXv2v)
        [[0 3 6]
         [2 4 6]]
        >>> vcvXv2v = vcv * Variable(value=((3,), (2,)))
        >>> print(vcvXv2v)
        [[0 3 6]
         [2 4 6]]
        >>> print(isinstance(vcvXv2v, CellVariable))
        1
        >>> print(vcvXv2v.rank)
        1
        >>> v2vXvcv = Variable(value=(3, 2)) * vcv
        >>> print(v2vXvcv)
        [[0 3 6]
         [2 4 6]]
        >>> v2vXvcv = Variable(value=((3,), (2,))) * vcv
        >>> print(v2vXvcv)
        [[0 3 6]
         [2 4 6]]
        >>> print(isinstance(v2vXvcv, CellVariable))
        1
        >>> print(v2vXvcv.rank)
        1

        >>> vcvXv3v = vcv * Variable(value=(3, 2, 1))
        >>> print(vcvXv3v)
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(vcvXv3v, CellVariable)
        1
        >>> print(vcvXv3v.rank)
        1
        >>> v3vXvcv = Variable(value=(3, 2, 1)) * vcv
        >>> print(v3vXvcv)
        [[0 2 2]
         [3 4 3]]
        >>> isinstance(v3vXvcv, CellVariable)
        1
        >>> print(v3vXvcv.rank)
        1

        >>> vcvXv4v = vcv * Variable(value=(3, 2, 1, 0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v4vXvcv = Variable(value=(3, 2, 1, 0)) * vcv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


    rank-1 `FaceVariable` * rank-1 `FaceVariable`

        >>> vfvXvfv = vfv * vfv
        >>> print(vfvXvfv)
        [[ 0  1  4  9  1  4  9 36  4  1]
         [ 1  4  9 16  9 16 25 81 36  9]]
        >>> isinstance(vfvXvfv, FaceVariable)
        1
        >>> print(vfvXvfv.rank)
        1

    rank-1 `FaceVariable` * Scalar

        >>> vfvXs = vfv * 3
        >>> print(vfvXs)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print(isinstance(vfvXs, FaceVariable))
        1
        >>> print(vfvXs.rank)
        1
        >>> sXvfv = 3 * vfv
        >>> print(sXvfv)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print(isinstance(sXvfv, FaceVariable))
        1
        >>> print(sXvfv.rank)
        1

    rank-1 `FaceVariable` * Vector

        >>> vfvXv2 = vfv * (3, 2)
        >>> print(vfvXv2)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print(isinstance(vfvXv2, FaceVariable))
        1
        >>> print(vfvXv2.rank)
        1
        >>> v2Xvfv = (3, 2) * vfv
        >>> print(v2Xvfv)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print(isinstance(v2Xvfv, FaceVariable))
        1
        >>> print(v2Xvfv.rank)
        1

        >>> vfvXv3 = vfv * (2, 1, 0) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3Xvfv = (2, 1, 0) * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


        >>> vfvXv10 = vfv * (9, 8, 7, 6, 5, 4, 3, 2, 1, 0)
        >>> print(vfvXv10)
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(vfvXv10, FaceVariable)
        1
        >>> print(vfvXv10.rank)
        1
        >>> v10Xvfv = (9, 8, 7, 6, 5, 4, 3, 2, 1, 0) * vfv
        >>> print(v10Xvfv)
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(v10Xvfv, FaceVariable)
        1
        >>> print(v10Xvfv.rank)
        1

    rank-1 `FaceVariable` * `Variable` Scalar

        >>> vfvXsv = vfv * Variable(value=3)
        >>> print(vfvXsv)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print(isinstance(vfvXsv, FaceVariable))
        1
        >>> print(vfvXsv.rank)
        1
        >>> svXvfv = Variable(value=3) * vfv
        >>> print(svXvfv)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 3  6  9 12  9 12 15 27 18  9]]
        >>> print(isinstance(svXvfv, FaceVariable))
        1
        >>> print(svXvfv.rank)
        1

    rank-1 `FaceVariable` * `Variable` Vector

        >>> vfvXv2v = vfv * Variable(value=(3, 2))
        >>> print(vfvXv2v)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print(isinstance(vfvXv2v, FaceVariable))
        1
        >>> print(vfvXv2v.rank)
        1
        >>> v2vXvfv = Variable(value=(3, 2)) * vfv
        >>> print(v2vXvfv)
        [[ 0  3  6  9  3  6  9 18  6  3]
         [ 2  4  6  8  6  8 10 18 12  6]]
        >>> print(isinstance(v2vXvfv, FaceVariable))
        1
        >>> print(v2vXvfv.rank)
        1

        >>> vfvXv3v = vfv * Variable(value=(2, 1, 0)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXvfv = Variable(value=(2, 1, 0)) * vfv #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


        >>> vfvXv10v = vfv * Variable(value=(9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
        >>> print(vfvXv10v)
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(vfvXv10v, FaceVariable)
        1
        >>> print(vfvXv10v.rank)
        1
        >>> v10vXvfv = Variable(value=(9, 8, 7, 6, 5, 4, 3, 2, 1, 0)) * vfv
        >>> print(v10vXvfv)
        [[ 0  8 14 18  5  8  9 12  2  0]
         [ 9 16 21 24 15 16 15 18  6  0]]
        >>> isinstance(v10vXvfv, FaceVariable)
        1
        >>> print(v10vXvfv.rank)
        1


    Scalar * `Variable` Scalar

        >>> sXsv = 3 * Variable(value=3)
        >>> print(sXsv)
        9
        >>> print(isinstance(sXsv, Variable))
        1
        >>> svXs = Variable(value=3) * 3
        >>> print(svXs)
        9
        >>> print(isinstance(svXs, Variable))
        1

    Scalar * `Variable` Vector

        >>> sXv2v = 3 * Variable(value=(3, 2))
        >>> print(sXv2v)
        [9 6]
        >>> print(isinstance(sXv2v, Variable))
        1
        >>> v2vXs = Variable(value=(3, 2)) * 3
        >>> print(v2vXs)
        [9 6]
        >>> print(isinstance(v2vXs, Variable))
        1



    Vector * `Variable` Scalar

        >>> vXsv = (3, 2) * Variable(value=3)
        >>> print(vXsv)
        [9 6]
        >>> print(isinstance(vXsv, Variable))
        1
        >>> svXv = Variable(value=3) * (3, 2)
        >>> print(svXv)
        [9 6]
        >>> print(isinstance(svXv, Variable))
        1

    Vector * `Variable` Vector

        >>> vXv2v = (3, 2) * Variable(value=(3, 2))
        >>> print(vXv2v)
        [9 4]
        >>> print(isinstance(vXv2v, Variable))
        1
        >>> v2vXv = Variable(value=(3, 2)) * (3, 2)
        >>> print(v2vXv)
        [9 4]
        >>> print(isinstance(v2vXv, Variable))
        1

        >>> vXv3v = (3, 2, 1) * Variable(value=(3, 2)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int
        >>> v3vXv = Variable(value=(3, 2)) * (3, 2, 1) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int


    `Variable` Scalar * `Variable` Scalar

        >>> svXsv = Variable(value=3) * Variable(value=3)
        >>> print(svXsv)
        9
        >>> print(isinstance(svXsv, Variable))
        1

    `Variable` Scalar * `Variable` Vector

        >>> svXv2v = Variable(value=3) * Variable(value=(3, 2))
        >>> print(svXv2v)
        [9 6]
        >>> print(isinstance(svXv2v, Variable))
        1
        >>> v2vXsv = Variable(value=(3, 2)) * Variable(value=3)
        >>> print(v2vXsv)
        [9 6]
        >>> print(isinstance(v2vXsv, Variable))
        1


    `Variable` Vector * `Variable` Vector

        >>> v2vXv2v = Variable(value=(3, 2)) * Variable(value=(3, 2))
        >>> print(v2vXv2v)
        [9 4]
        >>> print(isinstance(v2vXv2v, Variable))
        1

        >>> v3vXv2v = Variable(value=(3, 2, 1)) * Variable(value=(3, 2)) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
              ...
        TypeError: can't multiply sequence to non-int

    Test for weird bug that was appearing in inline. Caused by the intermediate
    operators not getting marked fresh.

        >>> class Alpha(Variable):
        ...     def __init__(self, var):
        ...         Variable.__init__(self)
        ...         self.var = self._requires(var)
        ...     def _calcValue(self):
        ...         return self.var.value

        >>> coeff = Variable()
        >>> alpha = Alpha(-coeff / 1)
        >>> print(numerix.allclose(alpha.value, 0.0))
        True
        >>> coeff.value = (-10.0)
        >>> print(numerix.allclose(alpha.value, 10))
        True
        >>> coeff.value = (10.0)
        >>> print(numerix.allclose(alpha.value, -10))
        True

    Test to prevent divide by zero evaluation before value is
    requested.  The request is caused by the Variable requiring
    its unit to see whether it can do an inline calculation in
    `_UnaryOperatorVariable()`.

        >>> T = Variable()
        >>> from fipy import numerix
        >>> v = numerix.exp(-T / (1. *  T))

    Following is a test case for an error when turning a `binOp` into an array

        >>> print(numerix.array(Variable(value=numerix.array([ 1.,])) * [ 1.,]))
        [ 1.]

    It seems that numpy's `__rmul__` coercion is very strange

        >>> type(numerix.array([1., 2.]) * Variable([1., 2.])) # doctest: +ELLIPSIS
        <class 'fipy.variables.binaryOperatorVariable...binOp'>

    Test inlining

        >>> v0 = Variable(numerix.ones(2, 'd'))
        >>> v1 = Variable(numerix.ones(2, 'd'))
        >>> v = v1 * v0
        >>> print(v)
        [ 1.  1.]
        >>> v0[1] = 0.5
        >>> print(v)
        [ 1.   0.5]

    Test inline indexing

        >>> mesh = Grid2D(nx=3, ny=3)
        >>> v1 = CellVariable(mesh=mesh, value=numerix.arange(9))
        >>> a = v1 * (1, -1)
        >>> print(a)
        [[ 0  1  2  3  4  5  6  7  8]
         [ 0 -1 -2 -3 -4 -5 -6 -7 -8]]
        >>> x, y = mesh.cellCenters
        >>> v1[(x == 0.5) & (y == 0.5)] = 0
        >>> print(a)
        [[ 0  1  2  3  4  5  6  7  8]
         [ 0 -1 -2 -3 -4 -5 -6 -7 -8]]

    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





"""Test numeric implementation of the mesh
"""
from __future__ import unicode_literals

__all__ = []

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(
        docTestModuleNames = (
            'fipy.variables.variable',
            'fipy.variables.meshVariable',
            'fipy.variables.cellVariable',
            'fipy.variables.faceVariable',
            'fipy.variables.operatorVariable',
            'fipy.variables.betaNoiseVariable',
            'fipy.variables.exponentialNoiseVariable',
            'fipy.variables.gammaNoiseVariable',
            'fipy.variables.gaussianNoiseVariable',
            'fipy.variables.uniformNoiseVariable',
            'fipy.variables.modularVariable',
            'fipy.variables.binaryOperatorVariable',
            'fipy.variables.unaryOperatorVariable',
            'fipy.variables.coupledCellVariable',
            'fipy.variables.cellToFaceVariable',
            'fipy.variables.faceGradVariable',
            'fipy.variables.gaussCellGradVariable',
            'fipy.variables.faceGradContributionsVariable',
            'fipy.variables.surfactantConvectionVariable',
            'fipy.variables.surfactantVariable',
            'fipy.variables.levelSetDiffusionVariable',
            'fipy.variables.distanceVariable'
        ))

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



from __future__ import unicode_literals
from builtins import range
__docformat__ = 'restructuredtext'

__all__ = []

import scipy.sparse as sp
from fipy.tools import numerix

from fipy.matrices.sparseMatrix import _SparseMatrix

class _ScipyMatrix(_SparseMatrix):

    """class wrapper for a scipy sparse matrix.

    `_ScipyMatrix` is always `NxN`.
    Allows basic python operations __add__, __sub__ etc.
    Facilitate matrix populating in an easy way.
    """

    def __init__(self, matrix):
        """Creates a `_ScipyMatrix`.

        Parameters
        ----------
        matrix : ~scipy.sparse.csr_matrix
            The internal SciPy matrix
        """
        self.matrix = matrix

    def getCoupledClass(self):
        return _CoupledScipyMeshMatrix

    def copy(self):
        return _ScipyMatrix(matrix=self.matrix.copy())

    def __getitem__(self, index):
        m = self.matrix[index]
        if isinstance(m, type(0)) or isinstance(m, type(0.)):
            return m
        else:
            return _ScipyMatrix(matrix=m)

    def __iadd__(self, other):
        return self._iadd(other)

    def _iadd(self, other, sign=1):
        if hasattr(other, "matrix"):
            self.matrix = self.matrix + (sign * other.matrix)
        elif type(other) in [float, int]:
            fillVec = numerix.repeat(other, self.matrix.nnz)

            self.matrix = self.matrix \
                          + sp.csr_matrix((fillVec, self.matrix.nonzero()),
                                          self.matrix.shape)
        else:
            self.matrix = self.matrix + (sign * other)

        return self

    def __add__(self, other):
        """
        Add two sparse matrices

            >>> L = _ScipyMatrixFromShape(size=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> print(L + _ScipyIdentityMatrix(size=3))
             1.000000  10.000000   3.000000  
                ---     4.141593      ---    
             2.500000      ---     1.000000  

            >>> print(L + 0)
                ---    10.000000   3.000000  
                ---     3.141593      ---    
             2.500000      ---        ---    

            >>> print(L + 3)
            Traceback (most recent call last):
            ...
            AttributeError: 'int' object has no attribute 'matrix'
        """

        if other == 0:
            return self
        else:
            return _ScipyMatrix(matrix=self.matrix + other.matrix)

    __radd__ = __add__

    def __sub__(self, other):

        if other == 0:
            return self
        else:
            L = self.matrix.copy()
            L -= other.matrix
            return _ScipyMatrix(matrix=L)

    def __rsub__(self, other):
        return -self + other

    def __isub__(self, other):
            return self._iadd(other, sign=-1)

    def __mul__(self, other):
        """
        Multiply a sparse matrix by another sparse matrix

        >>> L1 = _ScipyMatrixFromShape(size=3)
        >>> L1.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
        >>> L2 = _ScipyIdentityMatrix(size=3)
        >>> L2.put([4.38, 12357.2, 1.1], [2, 1, 0], [1, 0, 2])

        >>> tmp = numerix.array(((1.23572000e+05, 2.31400000e+01, 3.00000000e+00),
        ...                      (3.88212887e+04, 3.14159265e+00, 0.00000000e+00),
        ...                      (2.50000000e+00, 0.00000000e+00, 2.75000000e+00)))

        >>> numerix.allclose((L1 * L2).numpyArray, tmp)
        1

        or a sparse matrix by a vector

        >>> tmp = numerix.array((29., 6.28318531, 2.5))
        >>> numerix.allclose(L1 * numerix.array((1, 2, 3), 'd'), tmp)
        1

        or a vector by a sparse matrix

        >>> tmp = numerix.array((7.5, 16.28318531,  3.))
        >>> numerix.allclose(numerix.array((1, 2, 3), 'd') * L1, tmp) ## The multiplication is broken. Numpy is calling __rmul__ for every element instead of with  the whole array.
        1
        """
        N = self.matrix.shape[0]

        if isinstance(other, _ScipyMatrix):
            return _ScipyMatrix(matrix=(self.matrix * other.matrix))
        else:
            shape = numerix.shape(other)
            if shape == ():
                return _ScipyMatrix(matrix=(self.matrix * other))
            elif shape == (N,):
                return self.matrix * other
            else:
                raise TypeError

    def __rmul__(self, other):
        if isinstance(numerix.ones(1, 'l'), type(other)):
            y = self.matrix.transpose() * other.copy()
            return y
        else:
            return self * other

    def asformat(self, *args, **kwargs):
        return self.matrix.asformat(*args, **kwargs)

    @property
    def _shape(self):
        return self.matrix.shape

    @property
    def _range(self):
        return list(range(self._shape[1])), list(range(self._shape[0]))

    def put(self, vector, id1, id2):
        """
        Put elements of `vector` at positions of the matrix corresponding to (`id1`, `id2`)

            >>> L = _ScipyMatrixFromShape(size=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> print(L)
                ---    10.000000   3.000000  
                ---     3.141593      ---    
             2.500000      ---        ---    
        """
        assert(len(id1) == len(id2) == len(vector))

        # done in such a way to vectorize everything
        tempVec = numerix.array(vector) - self.matrix[id1, id2].flat
        tempMat = sp.csr_matrix((tempVec, (id1, id2)), self.matrix.shape)

        self.matrix = self.matrix + tempMat

    def putDiagonal(self, vector):
        """
        Put elements of `vector` along diagonal of matrix

            >>> L = _ScipyMatrixFromShape(size=3)
            >>> L.putDiagonal([3., 10., numerix.pi])
            >>> print(L)
             3.000000      ---        ---    
                ---    10.000000      ---    
                ---        ---     3.141593  
            >>> L.putDiagonal([10., 3.])
            >>> print(L)
            10.000000      ---        ---    
                ---     3.000000      ---    
                ---        ---     3.141593  
        """
        if type(vector) in [int, float]:
            vector = numerix.repeat(vector, self._shape[0])

        self.matrix.setdiag(vector)

    def take(self, id1, id2):
        return self.matrix[id1, id2]

    def takeDiagonal(self):
        return self.matrix.diagonal()

    def addAt(self, vector, id1, id2):
        """
        Add elements of `vector` to the positions in the matrix corresponding to (`id1`,`id2`)

            >>> L = _ScipyMatrixFromShape(size=3)
            >>> L.put([3., 10., numerix.pi, 2.5], [0, 0, 1, 2], [2, 1, 1, 0])
            >>> L.addAt([1.73, 2.2, 8.4, 3.9, 1.23], [1, 2, 0, 0, 1], [2, 2, 0, 0, 2])
            >>> print(L)
            12.300000  10.000000   3.000000  
                ---     3.141593   2.960000  
             2.500000      ---     2.200000  
        """
        assert(len(id1) == len(id2) == len(vector))

        temp = sp.csr_matrix((vector, (id1, id2)), self.matrix.shape)

        self.matrix = self.matrix + temp

    def addAtDiagonal(self, vector):
        if type(vector) in [type(1), type(1.)]:
            vector = numerix.repeat(vector, self._shape[0])

        ids = numerix.arange(len(vector))
        self.addAt(vector, ids, ids)

    @property
    def numpyArray(self):
        return self.matrix.toarray()

    def matvec(self, x):
        """This method is required for scipy solvers.
        """
        return self * x

    def exportMmf(self, filename):
        """Exports the matrix to a Matrix Market file of the given `filename`.
        """
        from scipy.io import mmio

        mmio.mmwrite(filename, self.matrix)

    def __getitem__(self, indices):
        return self.matrix[indices]

class _ScipyMatrixFromShape(_ScipyMatrix):

    def __init__(self, size, bandwidth=0, sizeHint=None, matrix=None, storeZeros=True):
        """Instantiates and wraps a scipy sparse matrix

        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The `Mesh` to assemble the matrix for.
        bandwidth : int
            The proposed band width of the matrix.
        storeZeros : bool
            Instructs scipy to store zero values if possible.
        """
        if matrix is None:
            matrix = sp.csr_matrix((size, size))

        _ScipyMatrix.__init__(self, matrix=matrix)

class _ScipyMeshMatrix(_ScipyMatrixFromShape):

    def __init__(self, mesh, bandwidth=0, sizeHint=None, matrix=None, numberOfVariables=1, numberOfEquations=1, storeZeros=True):

        """Creates a `_ScipyMatrixFromShape` associated with a `Mesh`.

        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The `Mesh` to assemble the matrix for.
        bandwidth : int
            The proposed band width of the matrix.
        numberOfVariables : int
            The columns of the matrix is determined by `numberOfVariables * self.mesh.numberOfCells`.
        numberOfEquations : int
            The rows of the matrix is determined by `numberOfEquations * self.mesh.numberOfCells`.
        storeZeros : bool
            Instructs scipy to store zero values if possible.
        """
        self.mesh = mesh
        self.numberOfVariables = numberOfVariables
        size = self.numberOfVariables * self.mesh.numberOfCells
        assert numberOfEquations == self.numberOfVariables
        _ScipyMatrixFromShape.__init__(self, size=size, matrix=matrix)

    def __mul__(self, other):
        if isinstance(other, _ScipyMeshMatrix):
            return _ScipyMeshMatrix(mesh=self.mesh,
                                    matrix=(self.matrix * other.matrix))
        else:
            return _ScipyMatrixFromShape.__mul__(self, other)

    def asTrilinosMeshMatrix(self):
        """Transforms a scipy matrix into a trilinos matrix and maintains the
        trilinos matrix as an attribute.

        Returns
        -------
        ~fipy.matrices.trilinosMatrix._TrilinosMatrix
        """
#         A = self.matrix.copy()
#         values, irow, jcol = A.find()
# 
#         if not hasattr(self, 'trilinosMatrix'):
#             if A.shape[0] == 0:
#                 bandwidth = 0
#             else:
#                 bandwidth = int(numerix.ceil(float(len(values)) / float(A.shape[0])))
#             from fipy.matrices.trilinosMatrix import _TrilinosMeshMatrixKeepStencil
#             self.trilinosMatrix = _TrilinosMeshMatrixKeepStencil(mesh=self.mesh, bandwidth=bandwidth, numberOfVariables=self.numberOfVariables)
# 
#         self.trilinosMatrix.addAt(values, irow, jcol)
#         self.trilinosMatrix.finalize()
# 
#         return self.trilinosMatrix

        raise NotImplementedError

    def flush(self):
        """
        Deletes the copy of the scipy matrix held and calls `self.trilinosMatrix.flush()` if necessary.
        """
#         if hasattr(self, 'trilinosMatrix'):
#             if hasattr(self.matrix, 'storeZeros'):
#                 self.trilinosMatrix.flush(cacheStencil=self.matrix.storeZeros)
#             else:
#                 self.trilinosMatrix.flush(cacheStencil=False)

        if (not hasattr(self, 'cache')) or (self.cache is False):
            del self.matrix

    def _test(self):
        """
        Tests

        >>> m = _ScipyMatrixFromShape(size=3, storeZeros=True)
        >>> m.addAt((1., 0., 2.), (0, 2, 1), (1, 2, 0))
        >>> nonZeroIdx = m.matrix.nonzero()
        >>> print(not hasattr(m.matrix, 'storeZeros') or numerix.allequal(nonZeroIdx, [(0, 1), (1, 0), (2, 2)]))
        True
        >>> print(not hasattr(m.matrix, 'storeZeros') or numerix.allequal(m.matrix[nonZeroIdx].toarray(), [1., 2., 0.]))
        True
        >>> m = _ScipyMatrixFromShape(size=3, storeZeros=False)
        >>> m.addAt((1., 0., 2.), (0, 2, 1), (1, 2, 0))
        >>> nonZeroIdx = m.matrix.nonzero()
        >>> print(numerix.allequal(nonZeroIdx, [(0, 1), (1, 0)]))
        True
        >>> print(numerix.allequal(numerix.array(m.matrix[nonZeroIdx]), numerix.array([1.0, 2.0])))
        True

        """
        pass

class _ScipyIdentityMatrix(_ScipyMatrixFromShape):
    """
    Represents a sparse identity matrix for scipy.
    """
    def __init__(self, size):
        """
        Create a sparse matrix with `1` in the diagonal

            >>> print(_ScipyIdentityMatrix(size=3))
             1.000000      ---        ---    
                ---     1.000000      ---    
                ---        ---     1.000000  
        """
        _ScipyMatrixFromShape.__init__(self, size=size, bandwidth = 1)
        ids = numerix.arange(size)
        self.put(numerix.ones(size, 'd'), ids, ids)

class _ScipyIdentityMeshMatrix(_ScipyIdentityMatrix):
    def __init__(self, mesh):
        """
        Create a sparse matrix associated with a `Mesh` with `1` in the diagonal

            >>> from fipy import Grid1D
            >>> from fipy.tools import serialComm
            >>> mesh = Grid1D(nx=3, communicator=serialComm)
            >>> print(_ScipyIdentityMeshMatrix(mesh=mesh))
             1.000000      ---        ---    
                ---     1.000000      ---    
                ---        ---     1.000000  
        """
        _ScipyIdentityMatrix.__init__(self, size=mesh.numberOfCells)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()


